using System.Collections.ObjectModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Reflection;
using System.Text.Json;
using System.Text.Json.Serialization;
using View_by_Distance.FaceRecognitionDotNet;
using View_by_Distance.Metadata.Models;
using View_by_Distance.Property.Models;
using View_by_Distance.Resize.Models;
using View_by_Distance.Shared.Models;
using View_by_Distance.Shared.Models.Properties;
using View_by_Distance.Shared.Models.Stateless;
using View_by_Distance.Shared.Models.Stateless.Methods;

namespace View_by_Distance.Face.Models;

/// <summary>
// List<D_Faces>
/// </summary>
public class D_Face : IFaceD
{

    protected readonly string _FileNameExtension;
    public string FileNameExtension => _FileNameExtension;

    protected readonly string _HiddenFileNameExtension;
    public string HiddenFileNameExtension => _HiddenFileNameExtension;

    private readonly Model _Model;
    private readonly string _ArgZero;
    private readonly bool _OverrideForFaceImages;
    private readonly bool _LoadPhotoPrismLocations;
    private readonly ImageCodecInfo _ImageCodecInfo;
    private readonly ModelParameter _ModelParameter;
    private readonly PredictorModel _PredictorModel;
    private readonly bool _CheckDFaceAndUpWriteDates;
    private readonly bool _PropertiesChangedForFaces;
    private readonly ConstructorInfo _ConstructorInfo;
    private readonly float _RectangleIntersectMinimum;
    private readonly int _FaceDistanceHiddenImageFactor;
    private readonly EncoderParameters _EncoderParameters;
    private readonly ImageCodecInfo _HiddenImageCodecInfo;
    private readonly Dictionary<string, string[]> _FileGroups;
    private readonly bool _ForceFaceLastWriteTimeToCreationTime;
    private readonly EncoderParameters _HiddenEncoderParameters;
    private readonly IPropertyConfiguration _PropertyConfiguration;
    private readonly JsonSerializerOptions _WriteIndentedAndWhenWritingNull;

    public D_Face(
        string argZero,
        IPropertyConfiguration propertyConfiguration,
        bool checkDFaceAndUpWriteDates,
        EncoderParameters encoderParameters,
        int faceDistanceHiddenImageFactor,
        string filenameExtension,
        bool forceFaceLastWriteTimeToCreationTime,
        EncoderParameters hiddenEncoderParameters,
        string hiddenFileNameExtension,
        ImageCodecInfo hiddenImageCodecInfo,
        ImageCodecInfo imageCodecInfo,
        bool loadPhotoPrismLocations,
        string modelDirectory,
        string modelName,
        bool overrideForFaceImages,
        string predictorModelName,
        bool propertiesChangedForFaces,
        float[] rectangleIntersectMinimums)
    {
        _ArgZero = argZero;
        _FileGroups = [];
        _ImageCodecInfo = imageCodecInfo;
        _EncoderParameters = encoderParameters;
        _FileNameExtension = filenameExtension;
        _HiddenImageCodecInfo = hiddenImageCodecInfo;
        _OverrideForFaceImages = overrideForFaceImages;
        _PropertyConfiguration = propertyConfiguration;
        _HiddenEncoderParameters = hiddenEncoderParameters;
        _HiddenFileNameExtension = hiddenFileNameExtension;
        _LoadPhotoPrismLocations = loadPhotoPrismLocations;
        _CheckDFaceAndUpWriteDates = checkDFaceAndUpWriteDates;
        _PropertiesChangedForFaces = propertiesChangedForFaces;
        _RectangleIntersectMinimum = rectangleIntersectMinimums.Min();
        _FaceDistanceHiddenImageFactor = faceDistanceHiddenImageFactor;
        _ForceFaceLastWriteTimeToCreationTime = forceFaceLastWriteTimeToCreationTime;
        (Model model, PredictorModel predictorModel, ModelParameter modelParameter) = GetModel(modelDirectory, modelName, predictorModelName);
        _Model = model;
        _PredictorModel = predictorModel;
        _ModelParameter = modelParameter;
        ConstructorInfo? constructorInfo = typeof(PropertyItem).GetConstructor(BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Public, null, [], null) ?? throw new Exception();
        _ConstructorInfo = constructorInfo;
        _WriteIndentedAndWhenWritingNull = new JsonSerializerOptions { WriteIndented = true, DefaultIgnoreCondition = JsonIgnoreCondition.WhenWritingNull };
    }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

    public void Update(string dResultsFullGroupDirectory)
    {
        _FileGroups.Clear();
        ReadOnlyDictionary<string, string[]> keyValuePairs = Shared.Models.Stateless.Methods.IPath.GetKeyValuePairs(_PropertyConfiguration, dResultsFullGroupDirectory, [_PropertyConfiguration.ResultCollection, _PropertyConfiguration.ResultContent]);
        foreach (KeyValuePair<string, string[]> keyValuePair in keyValuePairs)
            _FileGroups.Add(keyValuePair.Key, keyValuePair.Value);
    }

    private static (Model model, PredictorModel predictorModel, ModelParameter modelParameter) GetModel(string modelDirectory, string modelName, string predictorModelName)
    {
        (Model, PredictorModel, ModelParameter) result;
        Array array;
        Model? model = null;
        PredictorModel? predictorModel = null;
        array = Enum.GetValues(typeof(Model));
        foreach (Model check in array)
        {
            if (modelName.Contains(check.ToString()))
            {
                model = check;
                break;
            }
        }
        if (model is null)
            throw new Exception("Destination directory must have Model name!");
        model = model.Value;
        array = Enum.GetValues(typeof(PredictorModel));
        foreach (PredictorModel check in array)
        {
            if (predictorModelName.Contains(check.ToString()))
            {
                predictorModel = check;
                break;
            }
        }
        if (predictorModel is null)
            throw new Exception("Destination directory must have Predictor Model name!");
        predictorModel = predictorModel.Value;
        ModelParameter modelParameter = new()
        {
            CnnFaceDetectorModel = File.ReadAllBytes(Path.Combine(modelDirectory, "mmod_human_face_detector.dat")),
            FaceRecognitionModel = File.ReadAllBytes(Path.Combine(modelDirectory, "dlib_face_recognition_resnet_model_v1.dat")),
            PosePredictor5FaceLandmarksModel = File.ReadAllBytes(Path.Combine(modelDirectory, "shape_predictor_5_face_landmarks.dat")),
            PosePredictor68FaceLandmarksModel = File.ReadAllBytes(Path.Combine(modelDirectory, "shape_predictor_68_face_landmarks.dat"))
        };
        result = new(model.Value, predictorModel.Value, modelParameter);
        return result;
    }

#pragma warning disable CA1416

    private void SaveFaces(FileHolder resizedFileHolder, MetadataExtractor.GeoLocation? geoLocation, string? maker, string? model, List<(Shared.Models.Face, FileInfo?, string, bool)> collection)
    {
        int width;
        int height;
        Bitmap bitmap;
        short type = 2;
        FaceFile faceFile;
        Graphics graphics;
        Location? location;
        Rectangle rectangle;
        string faceFileJson;
        string faceEncodingJson;
        PropertyItem? propertyItem;
        using Bitmap source = new(resizedFileHolder.FullName);
        int artist = MetadataExtractor.Formats.Exif.ExifDirectoryBase.TagArtist;
        int userComment = MetadataExtractor.Formats.Exif.ExifDirectoryBase.TagUserComment;
        foreach ((Shared.Models.Face face, FileInfo? fileInfo, string fileName, bool save) in collection)
        {
            if (!save)
                continue;
            if (fileInfo is null)
                continue;
            if (face.FaceEncoding is null || face?.Location is null || face?.OutputResolution is null)
                continue;
            location = Shared.Models.Stateless.Methods.ILocation.GetLocation(face.Location, Shared.Models.Stateless.ILocation.Digits, Shared.Models.Stateless.ILocation.Factor, source.Height, source.Width, collection.Count);
            if (location is null)
                continue;
            width = location.Right - location.Left;
            height = location.Bottom - location.Top;
            faceEncodingJson = JsonSerializer.Serialize(face.FaceEncoding);
            rectangle = new Rectangle(location.Left, location.Top, width, height);
            faceFile = new(face.DateTime, geoLocation?.ToDmsString(), face.FaceParts, face.Location, maker, model, face.OutputResolution);
            faceFileJson = JsonSerializer.Serialize(faceFile, FaceFileGenerationContext.Default.FaceFile);
            using (bitmap = new(width, height))
            {
                using (graphics = Graphics.FromImage(bitmap))
                    graphics.DrawImage(source, new Rectangle(0, 0, width, height), rectangle, GraphicsUnit.Pixel);
                propertyItem = Property.Models.Stateless.IProperty.GetPropertyItem(_ConstructorInfo, artist, type, faceFileJson);
                bitmap.SetPropertyItem(propertyItem);
                propertyItem = Property.Models.Stateless.IProperty.GetPropertyItem(_ConstructorInfo, userComment, type, faceEncodingJson);
                bitmap.SetPropertyItem(propertyItem);
                bitmap.Save(fileInfo.FullName, _ImageCodecInfo, _EncoderParameters);
            }
            if (File.Exists(fileName))
                File.Delete(fileName);
            location = Shared.Models.Stateless.Methods.ILocation.GetLocation(_FaceDistanceHiddenImageFactor, face.Location, Shared.Models.Stateless.ILocation.Digits, Shared.Models.Stateless.ILocation.Factor, source.Height, source.Width, collection.Count);
            if (location is null)
                continue;
            width = location.Right - location.Left;
            height = location.Bottom - location.Top;
            rectangle = new Rectangle(location.Left, location.Top, width, height);
            using (bitmap = new(width, height))
            {
                using (graphics = Graphics.FromImage(bitmap))
                    graphics.DrawImage(source, new Rectangle(0, 0, width, height), rectangle, GraphicsUnit.Pixel);
                bitmap.Save(fileName, _HiddenImageCodecInfo, _HiddenEncoderParameters);
            }
            File.SetAttributes(fileName, FileAttributes.Hidden);
        }
    }

    private List<Shared.Models.Face> GetFaces(string outputResolution, Shared.Models.Property property, MappingFromItem mappingFromItem, Dictionary<string, int[]> outputResolutionToResize, List<Location> locations)
    {
        if (_PropertyConfiguration.NumberOfJitters is null)
            throw new NullReferenceException(nameof(_PropertyConfiguration.NumberOfJitters));
        if (_PropertyConfiguration.NumberOfTimesToUpsample is null)
            throw new NullReferenceException(nameof(_PropertyConfiguration.NumberOfTimesToUpsample));
        List<Shared.Models.Face> results = [];
        FaceRecognitionDotNet.Image? unknownImage;
        try
        { unknownImage = FaceRecognition.LoadImageFile(mappingFromItem.ResizedFileHolder.FullName); }
        catch (Exception)
        { unknownImage = null; }
        if (unknownImage is not null)
        {
            (int outputResolutionWidth, int outputResolutionHeight, int outputResolutionOrientation) = Resize.Models.Stateless.Methods.IResize.Get(outputResolution, outputResolutionToResize);
            List<(Location Location, FaceRecognitionDotNet.FaceEncoding? FaceEncoding, Dictionary<FacePart, FacePoint[]>? FaceParts)> collection;
            FaceRecognition faceRecognition = new(_PropertyConfiguration.NumberOfJitters.Value, _PropertyConfiguration.NumberOfTimesToUpsample.Value, _Model, _ModelParameter, _PredictorModel);
            collection = faceRecognition.GetCollection(unknownImage, locations, includeFaceEncoding: true, includeFaceParts: true);
            if (collection.Count == 0)
                results.Add(new(property, outputResolutionWidth, outputResolutionHeight, outputResolutionOrientation, location: null));
            else
            {
                double[] rawEncoding;
                Shared.Models.Face face;
                Shared.Models.FaceEncoding convertedFaceEncoding;
                foreach ((Location location, FaceRecognitionDotNet.FaceEncoding? faceEncoding, Dictionary<FacePart, FacePoint[]>? faceParts) in collection)
                {
                    face = new(property, outputResolutionWidth, outputResolutionHeight, outputResolutionOrientation, location);
                    if (faceEncoding is not null)
                    {
                        rawEncoding = faceEncoding.GetRawEncoding();
                        convertedFaceEncoding = new(rawEncoding, faceEncoding.Size);
                        face.SetFaceEncoding(convertedFaceEncoding);
                    }
                    if (faceParts is not null)
                        face.SetFaceParts(faceParts);
                    results.Add(face);
                }
            }
            unknownImage.Dispose();
            faceRecognition.Dispose();
        }
        return results;
    }

#pragma warning restore CA1416

    public List<Shared.Models.Face> GetFaces(string outputResolution, string dResultsFullGroupDirectory, FilePath filePath, List<Tuple<string, DateTime>> subFileTuples, List<string> parseExceptions, Shared.Models.Property property, MappingFromItem mappingFromItem, Dictionary<string, int[]> outputResolutionToResize, List<MappingFromPhotoPrism>? mappingFromPhotoPrismCollection)
    {
        List<Shared.Models.Face>? results;
        if (string.IsNullOrEmpty(dResultsFullGroupDirectory))
            throw new NullReferenceException(nameof(dResultsFullGroupDirectory));
        string? json;
        List<Location> locations;
        string[] changesFrom = [nameof(A_Property), nameof(B_Metadata), nameof(C_Resize)];
        List<DateTime> dateTimes = (from l in subFileTuples where changesFrom.Contains(l.Item1) select l.Item2).ToList();
        (_, int directoryIndex) = Shared.Models.Stateless.Methods.IPath.GetDirectoryNameAndIndex(_PropertyConfiguration, filePath);
        FileInfo fileInfo = new(Path.Combine(_FileGroups[_PropertyConfiguration.ResultCollection][directoryIndex], $"{mappingFromItem.FilePath.NameWithoutExtension}{mappingFromItem.FilePath.ExtensionLowered}.json"));
        if (_ForceFaceLastWriteTimeToCreationTime && !fileInfo.Exists && File.Exists(Path.ChangeExtension(fileInfo.FullName, ".delete")))
        {
            File.Move(Path.ChangeExtension(fileInfo.FullName, ".delete"), fileInfo.FullName);
            fileInfo.Refresh();
        }
        if (_ForceFaceLastWriteTimeToCreationTime && fileInfo.Exists && fileInfo.LastWriteTime != fileInfo.CreationTime)
        {
            File.SetLastWriteTime(fileInfo.FullName, fileInfo.CreationTime);
            fileInfo.Refresh();
        }
        if (_PropertiesChangedForFaces)
            results = null;
        else if (!fileInfo.Exists)
            results = null;
        else if (_CheckDFaceAndUpWriteDates && dateTimes.Count > 0 && dateTimes.Max() > fileInfo.LastWriteTime)
            results = null;
        else
        {
            json = Shared.Models.Stateless.Methods.IFace.GetJson(fileInfo.FullName);
            try
            {
                results = JsonSerializer.Deserialize<List<Shared.Models.Face>>(json);
                if (results is null)
                    throw new NullReferenceException(nameof(results));
                subFileTuples.Add(new Tuple<string, DateTime>(nameof(D_Face), fileInfo.LastWriteTime));
            }
            catch (Exception)
            {
                results = null;
                parseExceptions.Add(nameof(D_Face));
            }
        }
        if (!_LoadPhotoPrismLocations || mappingFromPhotoPrismCollection is null || results is null)
            locations = [];
        else
            locations = Shared.Models.Stateless.Methods.ILocation.GetLocations(results, mappingFromPhotoPrismCollection, _RectangleIntersectMinimum);
        if (results is null || locations.Count > 0)
        {
            results = GetFaces(outputResolution, property, mappingFromItem, outputResolutionToResize, locations);
            if (results.Count == 0)
                File.Move(mappingFromItem.ResizedFileHolder.FullName, $"{mappingFromItem.ResizedFileHolder.FullName}.err");
            else
            {
                bool updateDateWhenMatches = dateTimes.Count > 0 && fileInfo.Exists && dateTimes.Max() > fileInfo.LastWriteTime;
                DateTime? dateTime = !updateDateWhenMatches ? null : dateTimes.Max();
                json = JsonSerializer.Serialize(results, _WriteIndentedAndWhenWritingNull);
                if (Shared.Models.Stateless.Methods.IPath.WriteAllText(fileInfo.FullName, json, updateDateWhenMatches, compareBeforeWrite: true, updateToWhenMatches: dateTime))
                {
                    if (!_ForceFaceLastWriteTimeToCreationTime)
                        subFileTuples.Add(new Tuple<string, DateTime>(nameof(D_Face), DateTime.Now));
                    else
                    {
                        File.SetLastWriteTime(fileInfo.FullName, fileInfo.CreationTime);
                        fileInfo.Refresh();
                        subFileTuples.Add(new Tuple<string, DateTime>(nameof(D_Face), fileInfo.CreationTime));
                    }
                }
            }
        }
        return results;
    }

    public List<(Shared.Models.Face, FileInfo?, string, bool)> SaveFaces(string f, string dResultsFullGroupDirectory, FilePath filePath, List<Tuple<string, DateTime>> subFileTuples, List<string> parseExceptions, MappingFromItem mappingFromItem, ExifDirectory exifDirectory, List<Shared.Models.Face> faces)
    {
        List<(Shared.Models.Face, FileInfo?, string, bool Save)> results = [];
        bool save;
        FileInfo fileInfo;
        string deterministicHashCodeKey;
        string[] changesFrom = [nameof(A_Property), nameof(B_Metadata), nameof(C_Resize)];
        List<DateTime> dateTimes = (from l in subFileTuples where changesFrom.Contains(l.Item1) select l.Item2).ToList();
        (_, int directoryIndex) = Shared.Models.Stateless.Methods.IPath.GetDirectoryNameAndIndex(_PropertyConfiguration, filePath);
        string directory = Path.Combine(_FileGroups[_PropertyConfiguration.ResultContent][directoryIndex], mappingFromItem.FilePath.NameWithoutExtension);
        bool directoryExists = Directory.Exists(directory);
        foreach (Shared.Models.Face face in faces)
        {
            save = false;
            if (face.FaceEncoding is null || face.Location is null || face.OutputResolution is null)
            {
                results.Add(new(face, null, string.Empty, save));
                continue;
            }
            deterministicHashCodeKey = Shared.Models.Stateless.Methods.IMapping.GetDeterministicHashCodeKey(filePath, face.Location, Shared.Models.Stateless.ILocation.Digits, face.OutputResolution);
            fileInfo = new FileInfo(Path.Combine(directory, $"{deterministicHashCodeKey}{mappingFromItem.FilePath.ExtensionLowered}{_FileNameExtension}"));
            if (!directoryExists)
                save = true;
            else if (_OverrideForFaceImages)
                save = true;
            else if (!fileInfo.Exists)
                save = true;
            else if (_CheckDFaceAndUpWriteDates && dateTimes.Count > 0 && dateTimes.Max() > fileInfo.LastWriteTime)
                save = true;
            results.Add(new(face, fileInfo, Path.Combine(directory, $"{deterministicHashCodeKey}{mappingFromItem.FilePath.ExtensionLowered}{_HiddenFileNameExtension}"), save));
        }
        if (results.Any(l => l.Save))
        {
            if (!directoryExists)
                _ = Directory.CreateDirectory(directory);
            string? maker = Metadata.Models.Stateless.Methods.IMetadata.GetMaker(exifDirectory);
            string? model = Metadata.Models.Stateless.Methods.IMetadata.GetModel(exifDirectory);
            MetadataExtractor.GeoLocation? geoLocation = Metadata.Models.Stateless.Methods.IMetadata.GeoLocation(exifDirectory);
            SaveFaces(mappingFromItem.ResizedFileHolder, geoLocation, maker, model, results);
        }
        return results;
    }

#pragma warning disable CA1416

    void IFaceD.ReSaveFace(ExifDirectory exifDirectory, LocationContainer locationContainer, Shared.Models.Face face)
    {
        FileInfo fileInfo = new(locationContainer.FilePath.FullName);
        if (fileInfo.Exists)
        {
            short type = 2;
            string checkFile = $"{locationContainer.FilePath.FullName}.exif";
            int artist = MetadataExtractor.Formats.Exif.ExifDirectoryBase.TagArtist;
            string? maker = Metadata.Models.Stateless.Methods.IMetadata.GetMaker(exifDirectory);
            string? model = Metadata.Models.Stateless.Methods.IMetadata.GetModel(exifDirectory);
            MetadataExtractor.GeoLocation? geoLocation = Metadata.Models.Stateless.Methods.IMetadata.GeoLocation(exifDirectory);
            FaceFile faceFile = new(face.DateTime, geoLocation?.ToDmsString(), face.FaceParts, face.Location, maker, model, face.OutputResolution);
            string faceFileJson = JsonSerializer.Serialize(faceFile, FaceFileGenerationContext.Default.FaceFile);
            ConstructorInfo? constructorInfo = typeof(PropertyItem).GetConstructor(BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Public, null, [], null) ?? throw new Exception();
            PropertyItem? propertyItem = Property.Models.Stateless.IProperty.GetPropertyItem(constructorInfo, artist, type, faceFileJson);
            Bitmap bitmap = new(fileInfo.FullName);
            bitmap.SetPropertyItem(propertyItem);
            bitmap.Save(checkFile);
            bitmap.Dispose();
            File.SetLastWriteTime(checkFile, fileInfo.LastWriteTime);
            File.Delete(fileInfo.FullName);
            File.Move(checkFile, fileInfo.FullName);
        }
    }

#pragma warning restore CA1416

}